/***********************************************************************
 *
 *      OPPARSE.C
 *      Operand Parser for 68000 Assembler
 *
 *    Function: opParse()
 *      Parses an operand of the 68000 assembly language
 *      instruction and attempts to recognize its addressing
 *      mode. The p argument points to the string to be
 *      evaluated, and the function returns a pointer to the
 *      first character beyond the end of the operand.
 *      The function returns a description of the operands that
 *      it parses in an opDescriptor structure. The fields of
 *      the operand descriptor are filled in as appropriate for
 *      the mode of the operand that was found:
 *
 *       mode      returns the address mode (symbolic values
 *             defined in ASM.H)
 *       reg       returns the address or data register number
 *       data      returns the displacement or address or
 *             immediate value
 *       backRef   TRUE if data is the value of an expression
 *             that contains only constants and backwards
 *             references; FALSE otherwise.
 *       index     returns the index register
 *             (0-7 = D0-D7, 8-15 = A0-A7)
 *       size      returns the size to be used for the index
 *             register
 *
 *      The argument errorPtr is used to return an error code
 *      via the standard mechanism.
 *
 *   Usage: char *opParse(p, d, errorPtr)
 *      char *p;
 *      opDescriptor *d;
 *      int *errorPtr;
 *
 *      Author: Paul McKee
 *      ECE492    North Carolina State University
 *		Completely rewritten by Rudla Kudla 15.11.2000
 *
 *        Date: 10/10/86
 *
 *    Revision: 10/26/87
 *      Altered the immediate mode case to correctly flag
 *      constructs such as "#$1000(A5)" as syntax errors.
 *
 ************************************************************************/

#include "pila.h"
#include "asm.h"
#include "std.h"

#include "ComSupport.h"

#ifdef unix
    #include "strupr.h"
#endif

extern char gfPass2;


#define isTerm(c)   (isspace(c) || (c == ',') || c == '\0')
#define isRegNum(c) ((c >= '0') && (c <= '7'))


int ParseSize(char ** p, int * size) /*
Note:
	size argument is not modified, if there is no size specified.
	if there was some error, p argument is not modified


Syntax:
		size: "."  ["b" | "B" | "w" | "W" | "l" | "L" | val_size ]
		val_size: "[" <exp> "]"

Result:
	0 no size specified
	1 size was specified

	error code
*/{
	int err;
	char * s;
	int sz = 0;
	long v;
	char refPtr;
	valType type;

	s = *p;
	if (*s++=='.') {
		switch(*s++) {
		case 'B':
		case 'b':
			sz = BYTE; break;
		case 'w':
		case 'W':
			sz = WORD; 
			break;
		case 'l':
		case 'L':
			sz = LONG;
			break;
		case '[':
			s = eval(s, &type, &v, &refPtr, &err);

			if (err<=SEVERE) {
				if (v==BYTE || v==WORD || v==LONG) {
					if (*s==']') {
						s++;
						sz = (int)v;
					} else {
						err = 1;	// expecting ']' to close a[x] or d[x]
					}
				} else {
					err = 2;	// value for size must be 1, 2, or 4
				}
			} else {
				// printf("Severe error in expression.\n");
			}
			
			break;
		default:
			err = 4;	// Expected b, w, l or [
		}
	}
	if (sz!=0) {
		*size = sz;
		*p = s;
		return 1;
	} else {
		return 0;
	}
}


int	IsSymbolChar(char c)
{
	return isalnum(c) || c=='_' || c=='$'  || c == '?' || c == '@';
}

int	ParseSymbol(char ** s, char * name) /*
Purpose:
	Parse symbol name and store it into some help buffer.
	We suppose that buffer is long enough to accomodate the symbol.
Input:
	name	pointer tiop buffer into which the name will be stored
	s	start of symbol, it points after last symbol character if symbol was parsed
Result:
	Return 1 if symbol was found, 0 otherwise.
*/{
	char * p = *s;
	int err = 0;
	int i;

    // Symbol (symbols can start with an alphabetic char or '.', '_', '?', '$',
    // and include letters, numbers, '.', '_', '$', '?', '@')

    if (isalpha(*p) || *p == '.' || *p == '?' || *p == '_' || *p == '$') {
        /* Determine the value of a symbol */
        i = 0;
        /* Collect characters of the symbol's name
           (only SIGCHARS characters are significant) */
        do {
            if (i < SIGCHARS) {
                name[i++] = *p;
            }
            p++;
        } while (IsSymbolChar(*p));	//isalnum(*p) || *p == '_' || *p == '$'  || *p == '?' || *p == '@');

        name[i] = '\0';
		*s = p;
		return 1;
	}
	return 0;
}

int ParseReg(char ** p, int * regnum, int * err) /*
Purpose:
	Parse register id and size.
Syntax:
	SP		(=a7)
	a0-a7
	d0-d7
	d[n]	n in 0..7
	a[n]	n in 0..7
	r[n]	n in 0..15
	PC
Output:
	Error number (can be returned when some kind of error is encountered
	while evaluating some value.
Result:
	Return 1 if register was parsed, 0 otherwise.
*/ {
	char * s;
	int reg;
	int regbase = 0;	// it is set to 8 for address registers
	valType type;

	long data;
	char backRef;

	reg = -1;
	s = *p;

	if ( (s[0]=='S' || s[0] =='s') && (s[1]=='P' || s[1]=='p') && !IsSymbolChar(s[3])) {
		reg = 15;
		s+=2;

	} else if ((*s=='D' || *s=='d') || (*s=='A' || *s=='a')) {
		if (*s=='A' || *s=='a') {
			regbase = 8;
		}
		s++;
		if (*s>='0' && *s<='8' && !IsSymbolChar(s[1]) ) {
			reg = regbase + (*s-'0');
			s++;
		} else if (*s=='[') {
			s = eval(s+1, &type, &data, &backRef, err);
			if (*err<=SEVERE) {
				if (data>=0 && data<=7) {
					if (*s==']') {
						s++;
						reg = regbase+data;
					} else {
						return 3;	// expecting ']' to close a[x] or d[x]
					}
				} else {
					return 2;	// register number should evaluate to 0-7 for a[x] or d[x]
				}
			} else {
				// printf("Severe error in expression.\n");
			}
		}
	} else if ((*s=='R' || *s=='r') && (*s+1=='[')) {
		s = eval(s+2, &type, &data, &backRef, err);
		if (*err<=SEVERE) {
			if (data>=0 && data<=15) {
				if (*s==']') {
					s++;
					reg = regbase+data;
				} else {
					return 3;	// expecting ']' to close r[x]
				}
			} else {
				return 2;	// register number should evaluate to 0-7 for a[x] or d[x]
			}
		} else {
			// printf("Severe error in expression.\n");
		}
	} else if ( (s[0]=='P' || s[0] =='p') && (s[1]=='C' || s[1]=='c') && !IsSymbolChar(s[2]) ) {
		reg = PC;
		s+=2;
	}

	if (reg>=0 && reg<=16) {
		*regnum = reg;
		*p = s;
		return 1;
	}
	return 0;
}

int ParseRegAndSize(char ** p, int * reg, int * size, int * err) /*
*/ {
	int q;
	q = ParseReg(p, reg, err);
	if (q>0) {
		*size = ((*reg)<9)?WORD:LONG;			// default size for address registers is LONG, for data WORD
		ParseSize(p, size);
	}
	return q;
}


#define IsAn(x) (x>=8 && x<=15)


char *opParseIndirectOrPC(char *p, opDescriptor *d, int *errorPtr) /*
Note:
	It does not touch backref and data members.
	They are supposed to be initialized.
	Default values are backref=TRUE, data=0.
*/{
	int	reg;
	int	indexReg;
	int indexSize;
	char * s;
	int err = OK;

    if (p[0] == '(') {
		s = p+1;
		if (ParseReg(&s, &reg, &err)) {
			if (*s==')') {
				s++;
				if (*s=='+') {
					// Postincrement mode
					if (IsAn(reg)) {
						p = s+1;
						d->mode = AnIndPost;
						d->reg = reg-8;
						return p;
					} else {
						NEWERROR(*errorPtr, ADR_REG_EXPECTED);
						return NULL;
					}
				} else {
					if (reg==PC) {
						// PC relative (with displacement 0
						p = s;
						d->mode = PCDisp;
						return p;
					} else {
						// Address register indirect
						if (IsAn(reg)) {
							p = s;
							d->mode = AnInd;
							d->reg = reg-8;
							return p;
						} else {
							NEWERROR(*errorPtr, ADR_REG_EXPECTED);
							return NULL;
						}
					}
				}
			} else if (*s==',') {
				s++;
				s = skipSpace(s);
				indexSize = WORD;
				if (ParseRegAndSize(&s, &indexReg, &indexSize, &err)) {
					if (*s==')') {
						s++;
						if (indexSize == WORD || indexSize == LONG) {
							d->size = indexSize;
							if (reg==PC) {
								// PC relative with index (displacement = 0)
								p = s;
								d->mode = PCIndex;
								d->index = indexReg;

								return p;
							} else {
								if (IsAn(reg)) {
									// Address register indirect with index
									p = s;
									d->mode = AnIndIndex;
									d->reg = reg-8;
									d->index = indexReg;
									return p;
								} else {
									NEWERROR(*errorPtr, ADR_REG_EXPECTED);
									return NULL;
								}
							}

						} else {
							NEWERROR(*errorPtr, W_OR_L_SIZE_EXPECTED);
						}
					} else {
						// ')' expected
					}
				} else {
					// expected register
				}
			} else {
				// expected ')' or ','
			}

		}
	}
	return NULL;
}

void opPrintData(opDescriptor * d)
{
	if (d->backRef) {
		printf("%ld", d->data);
	} else {
		printf("?");
	}
}

void	opPrint(opDescriptor * d)
{
	switch(d->mode) {
	case DnDirect:
		printf("D%d",d->reg);
		break;
	case AnDirect:
		printf("A%d",d->reg);
		break;
	case AnInd:
		printf("(A%d)",d->reg);
		break;
	case AnIndPost:
		printf("(A%d)+",d->reg);
		break;
	case AnIndPre:
		printf("-(A%d)",d->reg);
		break;
	case AnIndDisp:
		opPrintData(d);
		printf("(A%d)",d->reg);
		break;
	case AnIndIndex:
		opPrintData(d);
		printf("(A%d,%c%d.%c)",d->reg, (d->index>=8)?'A':'D', (int)(d->index&7), (d->size==WORD)?'w':'l');
		break;
	case AbsShort:
		opPrintData(d);
		break;
	case AbsLong:
		opPrintData(d);
		break;
	case PCDisp:
		opPrintData(d);
		printf("(PC)");
		break;
	case PCIndex:
		printf("(PC,%c%d.%c)",d->reg, (int)(d->index&7), (d->size==WORD)?'w':'l');
		break;
	case Immediate:
		printf("#");
		opPrintData(d);
		break;
	case SRDirect:
		printf("SR");
		break;
	case CCRDirect:
		printf("CCR");
		break;
	case USPDirect:
		printf("USP");
		break;
	}
}

/*

char *opParseInner(char *p, opDescriptor *d, int *errorPtr);

char * opParse(char *p, opDescriptor *d, int *errorPtr)
 {
	char * res;
	res = opParseInner(p, d, errorPtr);
	if (*errorPtr < SEVERE) {
		printf("Line %d: ", gpsseCur->iLineNum);
		opPrint(d);
		printf(">>%s", res);
	}
	return res;
}
*/

char *opParse(char *p, opDescriptor *d, int *errorPtr) /*
Purpose:
	Parse instruction operand.

Syntax:

	Standard addressing modes:
  
	#constant

	An
	Dn

	(An)
	(An)+
	-(An)
	(An,Rn.w)		= 0(An,Rn.w)

	d(An)
	d(An,Rn.w)
	d(PC)			d must be label or constant
	d(PC,Rn.w)		d must be label or constant
	adr				constant adressing

Future syntax:
  
	Array-like addressing modes:

	An[value]		=  value(An)
	An[Rn]			=	(An,Rn)
	An[Rn+value]	=	value(An,Rn)
	
	If resulting type is label, we use PC relative addressing.

	label		=  label(PC)
	label[Rn]	=  label(PC,Rn)

	C-like syntax addresing modes:

	*An			=  (An)
	*An++		=  (An)+
	*--An		=	-(An)

	"#" el1
	"-(" el2 ")"
	el1 "(" el2 "," el3 ")" "+"

*/{

	char * s;
	int		reg;
	valType type;

	// Default values for data member
	
	d->data = 0;
	d->backRef = TRUE;

    /* Check for immediate mode */

    if (p[0] == '#') {
        s = eval(p+1, &type, &(d->data), &(d->backRef), errorPtr);

        /* If expression evaluates OK, then return */
        if (*errorPtr < SEVERE) {
            if (isTerm(*s)) {
                d->mode = Immediate;
                return s;
            } else {
                NEWERROR(*errorPtr, SYNTAX);
                return NULL;
            }
        } else {
            return NULL;
        }
    }

    /* Check for address or data register direct */

	s = p;
	if (ParseReg(&s, &reg, errorPtr)>0) {
		if (reg<8) {
			d->mode = DnDirect;
			d->reg = reg;
		} else if (reg<16) {
			d->mode = AnDirect;
			d->reg = reg-8;
		}
		return s;
	}

    /* Check for address register indirect with predecrement */

    if (p[0]=='-' && p[1] == '(' ) {
		s = p+2;
		if (ParseReg(&s, &reg, errorPtr)) {
			if (*s=')') {
				if (IsAn(reg)) {
					p = s+1;
					d->mode = AnIndPre;
					d->reg = reg-8;
					return p;
				} else {
					NEWERROR(*errorPtr, ADR_REG_EXPECTED);
					return NULL;
				}
			}
		}
    }

	/* Check for indirect modes */

	s = opParseIndirectOrPC(p, d, errorPtr);
	if (s!=NULL) {
		p = s;
		return p;
	}

    /* Check for Status Register direct */
    if (p[0] == 'S' && p[1] == 'R' && isTerm(p[2])) {
        d->mode = SRDirect;
        return p+2;
    }
    /* Check for Condition Code Register direct */
    if (p[0] == 'C' && p[1] == 'C' && p[2] == 'R' && isTerm(p[3])) {
        d->mode = CCRDirect;
        return p+3;
    }
    /* Check for User Stack Pointer direct */
    if (p[0] == 'U' && p[1] == 'S' && p[2] == 'P' && isTerm(p[3])) {
        d->mode = USPDirect;
        return p+3;
    }
    /* Check for Source Function Code register direct (68010) */
    if (p[0] == 'S' && p[1] == 'F' && p[2] == 'C' && isTerm(p[3])) {
        d->mode = SFCDirect;
        return p+3;
    }
    /* Check for Destination Function Code register direct (68010) */
    if (p[0] == 'D' && p[1] == 'F' && p[2] == 'C' && isTerm(p[3])) {
        d->mode = DFCDirect;
        return p+3;
    }
    /* Check for Vector Base Register direct (68010) */
    if (p[0] == 'V' && p[1] == 'B' && p[2] == 'R' && isTerm(p[3])) {
        d->mode = VBRDirect;
        return p+3;
    }


	// Check for immediate mode or indirect with offset

    s = eval(p, &type, &(d->data), &(d->backRef), errorPtr);
    if (*errorPtr < SEVERE) {
        // Check for absolute
        if (isTerm(s[0])) {
            // Determine size of absolute address (must be long if
            // the symbol isn't defined or if the value is too big.

            if (!d->backRef || d->data > 32767 || d->data < -32768) {
                d->mode = AbsLong;
            } else {
                d->mode = AbsShort;
            }
            return s;
        } else {

			s = opParseIndirectOrPC(s, d, errorPtr);

			if (*errorPtr < SEVERE) {

				switch(d->mode) {
				case AnInd:
					d->mode = AnIndDisp;
					break;
				case AnIndIndex:
				case PCDisp:
				case PCIndex:
					// these three remain same (index is just non zero)
					break;
				case AnIndPost:
					// if there is this mode, it is an error (can't have both displacement and postincrement)
					NEWERROR(*errorPtr, SYNTAX);
					return NULL;
				}
				p = s;
				return p;
			}
		}
	}

    /* If the operand doesn't match any pattern, return an error status */
    NEWERROR(*errorPtr, SYNTAX);
    return NULL;
}


